package xyz.lychee.lagfixer.libs.adventure.platform.viaversion;

import com.viaversion.viaversion.api.Via;
import com.viaversion.viaversion.api.connection.UserConnection;
import com.viaversion.viaversion.api.protocol.Protocol;
import com.viaversion.viaversion.api.protocol.packet.ClientboundPacketType;
import com.viaversion.viaversion.api.protocol.packet.PacketWrapper;
import com.viaversion.viaversion.api.type.Type;
import com.viaversion.viaversion.libs.gson.JsonElement;
import com.viaversion.viaversion.libs.gson.JsonParser;
import io.netty.buffer.ByteBuf;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.function.Consumer;
import java.util.function.Function;
import xyz.lychee.lagfixer.libs.adventure.audience.MessageType;
import xyz.lychee.lagfixer.libs.adventure.identity.Identity;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.Facet;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.FacetBase;
import xyz.lychee.lagfixer.libs.adventure.platform.facet.Knob;
import xyz.lychee.lagfixer.libs.adventure.text.Component;
import xyz.lychee.lagfixer.libs.adventure.text.serializer.gson.GsonComponentSerializer;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;

public class ViaFacet<V> extends FacetBase<V> implements Facet.Message<V, String> {
   private static final String PACKAGE = "com.viaversion.viaversion";
   private static final int SUPPORTED_VIA_MAJOR_VERSION = 4;
   private static final boolean SUPPORTED;
   private final Function<V, UserConnection> connectionFunction;
   private final int minProtocol;

   public ViaFacet(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction, final int minProtocol) {
      super(viewerClass);
      this.connectionFunction = connectionFunction;
      this.minProtocol = minProtocol;
   }

   public boolean isSupported() {
      return super.isSupported() && SUPPORTED && this.connectionFunction != null && this.minProtocol >= 0;
   }

   public boolean isApplicable(final @NotNull V viewer) {
      return super.isApplicable(viewer) && this.minProtocol > Via.getAPI().getServerVersion().lowestSupportedVersion() && this.findProtocol(viewer) >= this.minProtocol;
   }

   public @Nullable UserConnection findConnection(final @NotNull V viewer) {
      return (UserConnection)this.connectionFunction.apply(viewer);
   }

   public int findProtocol(final @NotNull V viewer) {
      UserConnection connection = this.findConnection(viewer);
      return connection != null ? connection.getProtocolInfo().getProtocolVersion() : -1;
   }

   public @NotNull String createMessage(final @NotNull V viewer, final @NotNull Component message) {
      int protocol = this.findProtocol(viewer);
      return protocol >= 713 ? (String)GsonComponentSerializer.gson().serialize(message) : (String)GsonComponentSerializer.colorDownsamplingGson().serialize(message);
   }

   static {
      boolean supported = false;

      try {
         Class.forName("com.viaversion.viaversion.api.ViaAPI").getDeclaredMethod("majorVersion");
         supported = Via.getAPI().majorVersion() == 4;
      } catch (Throwable var2) {
      }

      SUPPORTED = supported && Knob.isEnabled("viaversion", true);
   }

   public static final class TabList<V> extends ProtocolBased<V> implements Facet.TabList<V, String> {
      public TabList(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> userConnection) {
         super("1_16", "1_15_2", 713, "TAB_LIST", viewerClass, userConnection);
      }

      public void send(final V viewer, final @Nullable String header, final @Nullable String footer) {
         PacketWrapper packet = this.createPacket(viewer);
         packet.write(Type.COMPONENT, this.parse(header));
         packet.write(Type.COMPONENT, this.parse(footer));
         this.sendPacket(packet);
      }
   }

   public static final class BossBar<V> extends ProtocolBased<V> implements Facet.BossBarPacket<V> {
      private final Set<V> viewers;
      private UUID id;
      private String title;
      private float health;
      private int color;
      private int overlay;
      private byte flags;

      private BossBar(final @NotNull String fromProtocol, final @NotNull String toProtocol, final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction, final Collection<V> viewers) {
         super(fromProtocol, toProtocol, 356, "BOSSBAR", viewerClass, connectionFunction);
         this.viewers = new CopyOnWriteArraySet(viewers);
      }

      public void bossBarInitialized(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar) {
         super.bossBarInitialized(bar);
         this.id = UUID.randomUUID();
         this.broadcastPacket(0);
      }

      public void bossBarNameChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final @NotNull Component oldName, final @NotNull Component newName) {
         if (!this.viewers.isEmpty()) {
            this.title = this.createMessage(this.viewers.iterator().next(), newName);
            this.broadcastPacket(3);
         }

      }

      public void bossBarProgressChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final float oldPercent, final float newPercent) {
         this.health = newPercent;
         this.broadcastPacket(2);
      }

      public void bossBarColorChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Color oldColor, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Color newColor) {
         this.color = this.createColor(newColor);
         this.broadcastPacket(4);
      }

      public void bossBarOverlayChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Overlay oldOverlay, final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Overlay newOverlay) {
         this.overlay = this.createOverlay(newOverlay);
         this.broadcastPacket(4);
      }

      public void bossBarFlagsChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final @NotNull Set<xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag> flagsAdded, final @NotNull Set<xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag> flagsRemoved) {
         this.flags = this.createFlag(this.flags, flagsAdded, flagsRemoved);
         this.broadcastPacket(5);
      }

      public void sendPacket(final @NotNull V viewer, final int action) {
         PacketWrapper packet = this.createPacket(viewer);
         packet.write(Type.UUID, this.id);
         packet.write(Type.VAR_INT, action);
         if (action == 0 || action == 3) {
            packet.write(Type.COMPONENT, this.parse(this.title));
         }

         if (action == 0 || action == 2) {
            packet.write(Type.FLOAT, this.health);
         }

         if (action == 0 || action == 4) {
            packet.write(Type.VAR_INT, this.color);
            packet.write(Type.VAR_INT, this.overlay);
         }

         if (action == 0 || action == 5) {
            packet.write(Type.BYTE, this.flags);
         }

         this.sendPacket(packet);
      }

      public void broadcastPacket(final int action) {
         if (!this.isEmpty()) {
            Iterator var2 = this.viewers.iterator();

            while(var2.hasNext()) {
               V viewer = var2.next();
               this.sendPacket(viewer, action);
            }

         }
      }

      public void addViewer(final @NotNull V viewer) {
         if (this.viewers.add(viewer)) {
            this.sendPacket(viewer, 0);
         }

      }

      public void removeViewer(final @NotNull V viewer) {
         if (this.viewers.remove(viewer)) {
            this.sendPacket(viewer, 1);
         }

      }

      public boolean isEmpty() {
         return this.id == null || this.viewers.isEmpty();
      }

      public void close() {
         this.broadcastPacket(1);
         this.viewers.clear();
      }

      // $FF: synthetic method
      BossBar(String x0, String x1, Class x2, Function x3, Collection x4, Object x5) {
         this(x0, x1, x2, x3, x4);
      }

      public static class Builder1_9_To_1_15<V> extends ViaFacet<V> implements Facet.Builder<V, Facet.BossBar<V>> {
         public Builder1_9_To_1_15(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
            super(viewerClass, connectionFunction, 356);
         }

         public Facet.@NotNull BossBar<V> createBossBar(final @NotNull Collection<V> viewer) {
            return new BossBar("1_9", "1_8", this.viewerClass, this::findConnection, viewer);
         }
      }

      public static class Builder<V> extends ViaFacet<V> implements Facet.Builder<V, Facet.BossBar<V>> {
         public Builder(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
            super(viewerClass, connectionFunction, 713);
         }

         public Facet.@NotNull BossBar<V> createBossBar(final @NotNull Collection<V> viewer) {
            return new BossBar("1_16", "1_15_2", this.viewerClass, this::findConnection, viewer);
         }
      }
   }

   public static class Title<V> extends ProtocolBased<V> implements Facet.TitlePacket<V, String, List<Consumer<PacketWrapper>>, Consumer<V>> {
      protected Title(final @NotNull String fromProtocol, final @NotNull String toProtocol, final int minProtocol, final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
         super(fromProtocol, toProtocol, minProtocol, "TITLE", viewerClass, connectionFunction);
      }

      public Title(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
         this("1_16", "1_15_2", 713, viewerClass, connectionFunction);
      }

      public @NotNull List<Consumer<PacketWrapper>> createTitleCollection() {
         return new ArrayList();
      }

      public void contributeTitle(final @NotNull List<Consumer<PacketWrapper>> coll, final @NotNull String title) {
         coll.add((packet) -> {
            packet.write(Type.VAR_INT, 0);
            packet.write(Type.COMPONENT, this.parse(title));
         });
      }

      public void contributeSubtitle(final @NotNull List<Consumer<PacketWrapper>> coll, final @NotNull String subtitle) {
         coll.add((packet) -> {
            packet.write(Type.VAR_INT, 1);
            packet.write(Type.COMPONENT, this.parse(subtitle));
         });
      }

      public void contributeTimes(final @NotNull List<Consumer<PacketWrapper>> coll, final int inTicks, final int stayTicks, final int outTicks) {
         coll.add((packet) -> {
            packet.write(Type.VAR_INT, 3);
            packet.write(Type.INT, inTicks);
            packet.write(Type.INT, stayTicks);
            packet.write(Type.INT, outTicks);
         });
      }

      public @Nullable Consumer<V> completeTitle(final @NotNull List<Consumer<PacketWrapper>> coll) {
         return (v) -> {
            int i = 0;

            for(int length = coll.size(); i < length; ++i) {
               PacketWrapper pkt = this.createPacket(v);
               ((Consumer)coll.get(i)).accept(pkt);
               this.sendPacket(pkt);
            }

         };
      }

      public void showTitle(final @NotNull V viewer, final @NotNull Consumer<V> title) {
         title.accept(viewer);
      }

      public void clearTitle(final @NotNull V viewer) {
         PacketWrapper packet = this.createPacket(viewer);
         packet.write(Type.VAR_INT, 4);
         this.sendPacket(packet);
      }

      public void resetTitle(final @NotNull V viewer) {
         PacketWrapper packet = this.createPacket(viewer);
         packet.write(Type.VAR_INT, 5);
         this.sendPacket(packet);
      }
   }

   public static class ActionBarTitle<V> extends ProtocolBased<V> implements Facet.ActionBar<V, String> {
      public ActionBarTitle(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
         super("1_11", "1_10", 310, "TITLE", viewerClass, connectionFunction);
      }

      public void sendMessage(final @NotNull V viewer, final @NotNull String message) {
         PacketWrapper packet = this.createPacket(viewer);
         packet.write(Type.VAR_INT, 2);
         packet.write(Type.COMPONENT, this.parse(message));
         this.sendPacket(packet);
      }
   }

   public static class ActionBar<V> extends Chat<V> implements Facet.ActionBar<V, String> {
      public ActionBar(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
         super(viewerClass, connectionFunction);
      }

      public byte createMessageType(final @NotNull MessageType type) {
         return 2;
      }

      public void sendMessage(final @NotNull V viewer, final @NotNull String message) {
         this.sendMessage(viewer, Identity.nil(), message, MessageType.CHAT);
      }
   }

   public static class Chat<V> extends ProtocolBased<V> implements Facet.ChatPacket<V, String> {
      public Chat(final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
         super("1_16", "1_15_2", 713, "CHAT_MESSAGE", viewerClass, connectionFunction);
      }

      public void sendMessage(final @NotNull V viewer, final @NotNull Identity source, final @NotNull String message, final @NotNull Object type) {
         PacketWrapper packet = this.createPacket(viewer);
         packet.write(Type.COMPONENT, this.parse(message));
         packet.write(Type.BYTE, this.createMessageType(type instanceof MessageType ? (MessageType)type : MessageType.SYSTEM));
         packet.write(Type.UUID, source.uuid());
         this.sendPacket(packet);
      }
   }

   public static class ProtocolBased<V> extends ViaFacet<V> {
      private final Class<? extends Protocol<?, ?, ?, ?>> protocolClass;
      private final Class<? extends ClientboundPacketType> packetClass;
      private final int packetId;

      protected ProtocolBased(final @NotNull String fromProtocol, final @NotNull String toProtocol, final int minProtocol, final @NotNull String packetName, final @NotNull Class<? extends V> viewerClass, final @NotNull Function<V, UserConnection> connectionFunction) {
         super(viewerClass, connectionFunction, minProtocol);
         String protocolClassName = MessageFormat.format("{0}.protocols.protocol{1}to{2}.Protocol{1}To{2}", "com.viaversion.viaversion", fromProtocol, toProtocol);
         String packetClassName = MessageFormat.format("{0}.protocols.protocol{1}to{2}.ClientboundPackets{1}", "com.viaversion.viaversion", fromProtocol, toProtocol);
         Class<? extends Protocol<?, ?, ?, ?>> protocolClass = null;
         Class<? extends ClientboundPacketType> packetClass = null;
         int packetId = -1;

         try {
            protocolClass = Class.forName(protocolClassName);
            packetClass = Class.forName(packetClassName);
            ClientboundPacketType[] var12 = (ClientboundPacketType[])packetClass.getEnumConstants();
            int var13 = var12.length;

            for(int var14 = 0; var14 < var13; ++var14) {
               ClientboundPacketType type = var12[var14];
               if (type.getName().equals(packetName)) {
                  packetId = type.getId();
                  break;
               }
            }
         } catch (Throwable var16) {
         }

         this.protocolClass = protocolClass;
         this.packetClass = packetClass;
         this.packetId = packetId;
      }

      public boolean isSupported() {
         return super.isSupported() && this.protocolClass != null && this.packetClass != null && this.packetId >= 0;
      }

      public PacketWrapper createPacket(final @NotNull V viewer) {
         return PacketWrapper.create(this.packetId, (ByteBuf)null, this.findConnection(viewer));
      }

      public void sendPacket(final @NotNull PacketWrapper packet) {
         if (packet.user() != null) {
            try {
               packet.scheduleSend(this.protocolClass);
            } catch (Throwable var3) {
               Throwable error = var3;
               Knob.logError(error, "Failed to send ViaVersion packet: %s %s", new Object[]{packet.user(), packet});
            }

         }
      }

      public @NotNull JsonElement parse(final @NotNull String message) {
         return JsonParser.parseString(message);
      }
   }
}
